@@ -1,36 +1,35 @@
 #A* -------------------------------------------------------------------
 #B* This file contains source code for the PyMOL computer program
-#C* Copyright (c) Schrodinger, LLC. 
+#C* Copyright (c) Schrodinger, LLC.
 #D* -------------------------------------------------------------------
 #E* It is unlawful to modify or remove this copyright notice.
 #F* -------------------------------------------------------------------
-#G* Please see the accompanying LICENSE file for further information. 
+#G* Please see the accompanying LICENSE file for further information.
 #H* -------------------------------------------------------------------
 #I* Additional authors of this source file include:
-#-* 
-#-* 
+#-*
+#-*
 #-*
 #Z* -------------------------------------------------------------------
 
 from __future__ import print_function, absolute_import
 
-if __name__=='pymol.exporting':
+if True:
     import os
     import sys
     if sys.version_info[0] == 2:
         import thread
-        import cPickle
     else:
         import _thread as thread
-        import pickle as cPickle
     from . import selector
     import re
     import copy
-    
+
     import pymol
     cmd = sys.modules["pymol.cmd"]
     from .cmd import _cmd,lock,unlock,Shortcut,QuietException
     from chempy import io
+    from chempy.pkl import cPickle
     from chempy.sdf import SDF,SDFRec
     from .cmd import _feedback,fb_module,fb_mask, \
                      DEFAULT_ERROR, DEFAULT_SUCCESS, _raising, is_ok, is_error, \
@@ -39,11 +38,11 @@ if __name__=='pymol.exporting':
 
     def copy_image(quiet=1,_self=cmd): # incentive feature / proprietary
         r = DEFAULT_ERROR
-        if thread.get_ident() == pymol.glutThread:
+        if _self.is_gui_thread():
             r = _self._copy_image(_self,int(quiet))
         else:
-            r = _self.do('cmd._copy_image(quiet=%d,_self=cmd)'%int(quiet))
-        if _self._raising(r,_self): raise QuietException         
+            r = _self.do('cmd._copy_image(quiet=%d)'%int(quiet))
+        if _self._raising(r,_self): raise QuietException
         return r
 
     cache_action_dict = {
@@ -92,7 +91,7 @@ PYMOL API
     cmd.cache(string action, string scenes, int state, int quiet)
 
     '''
-        
+
         r = DEFAULT_ERROR
         action = cache_action_dict[cache_action_sc.auto_err(str(action),'action')]
         quiet = int(quiet)
@@ -111,7 +110,7 @@ PYMOL API
             cache_max = int(_self.get('cache_max'))
             if cache_max>0:
                 # allow double memory for an optimized cache
-                _self.set('cache_max',cache_max*2) 
+                _self.set('cache_max',cache_max*2)
             scenes = str(scenes)
             scene_list = scenes.split()
             cache_mode = int(_self.get('cache_mode'))
@@ -122,33 +121,33 @@ PYMOL API
                 scene = scene.strip()
                 if not quiet:
                     print(" cache: optimizing scene '%s'."%scene)
-                cmd.scene(scene,animate=0)
-                cmd.rebuild()                
-                cmd.refresh()
+                _self.scene(scene,animate=0)
+                _self.rebuild()
+                _self.refresh()
             if len(cur_scene):
-                cmd.scene(cur_scene,animate=0)
+                _self.scene(cur_scene,animate=0)
             else:
                 scene_list = _self.get_scene_list()
                 if len(scene_list):
-                    cmd.scene(scene_list[0],animate=0)
+                    _self.scene(scene_list[0],animate=0)
                 else:
                     if not quiet:
                         print(" cache: no scenes defined -- optimizing current display.")
-                    cmd.rebuild() 
-                    cmd.refresh()
+                    _self.rebuild()
+                    _self.refresh()
             usage = _self._cache_purge(-1,_self=_self)
             if cache_mode:
                 _self.set('cache_mode',cache_mode)
             else:
                 _self.set('cache_mode',2) # hmm... could use 1 here instead.
             _self.set('cache_max',cache_max) # restore previous limits
-            if not quiet:            
+            if not quiet:
                 print(" cache: optimization complete (~%0.1f MB)."%(usage*4/1000000.0))
         try:
             _self.lock(_self)
         finally:
             _self.unlock(r,_self)
-        if _self._raising(r,_self): raise QuietException         
+        if _self._raising(r,_self): raise QuietException
         return r
 
     _resn_to_aa =  {
@@ -172,33 +171,69 @@ PYMOL API
             'VAL' : 'V',
             'TRP' : 'W',
             'TYR' : 'Y',
+            # RNA
+            'A'   : 'A',
+            'U'   : 'U',
+            'G'   : 'G',
+            'C'   : 'C',
+            # DNA
+            'DA'  : 'A',
+            'DT'  : 'T',
+            'DG'  : 'G',
+            'DC'  : 'C',
             }
-    
-    def get_fastastr(selection="all", state=-1, quiet=1, _self=cmd):
-        dict = { 'seq' : {} }
-        # we use (alt '' or alt 'A') because 'guide' picks up 
-        # non-canonical structures: eg, 1ejg has residue 22 as a SER and 
-        # PRO, which guide will report twice
-        _self.iterate("("+selection+") and polymer and name CA and alt +A",
-                    "seq[model]=seq.get(model,[]);seq[model].append(resn)",space=dict)
-        seq = dict['seq']
-        result = []
-        for obj in _self.get_names("objects",selection='('+selection+')'):
-            if obj in seq:
-                cur_seq = [_resn_to_aa.get(x,'?') for x in seq[obj]]
-                result.append(">%s"%obj)
-                cur_seq = ''.join(cur_seq)
-                while len(cur_seq):
-                    if len(cur_seq)>=70:
-                        result.append(cur_seq[0:70])
-                        cur_seq=cur_seq[70:]
-                    else:
-                        result.append(cur_seq)
-                        break
-        result = '\n'.join(result)
-        if len(result):
-            result = result + '\n'
-        return result
+
+    def get_fastastr(selection="all", state=-1, quiet=1, key='', _self=cmd):
+        '''
+DESCRIPTION
+
+    API only. Get protein and nucleic acid sequences in fasta format.
+
+    Used for saving:
+    PyMOL> save foo.fasta
+
+    New in PyMOL 2.2:
+    - chain specific keys (key argument)
+    - nucleic acid support
+
+ARGUMENTS
+
+    selection = str: atom selection (reduced to "guide & alt +A") {default: all}
+
+    state = int: (only used if state > 0)
+
+    quiet = 0/1: UNUSED
+
+    key = str: python expression {default: model + "_" + chain}
+    Use key=model to get the old (non-chain specific) behavior
+        '''
+        import textwrap
+        import collections
+
+        seq = collections.OrderedDict()
+        lines = []
+
+        selection = "(" + selection + ") & guide & alt +A"
+        if int(state) > 0:
+            selection += ' & state {}'.format(state)
+
+        if not key:
+            key = 'model + "_" + chain'
+            # for discrete objects: state specific keys
+            key += ' + (":{}".format(state) if state else "")'
+
+        _self.iterate(selection,
+            "seq.setdefault((" + key + "),[]).append(resn)",
+            space={'seq': seq, 'str': str})
+
+        for key, resn_list in seq.items():
+            cur_seq = ''.join(_resn_to_aa.get(x, '?') for x in resn_list)
+            lines.append(">{}".format(key))
+            lines.extend(textwrap.wrap(cur_seq, 70))
+
+        if lines:
+            lines.append('')  # final newline
+        return '\n'.join(lines)
 
     def get_pdbstr(selection="all", state=-1, ref='', ref_state=-1, quiet=1, _self=cmd):
         '''
@@ -238,7 +273,7 @@ NOTES
         else:
             result = str(obj)
         return result
-    
+
     def _session_convert_legacy(session, version, _self=cmd):
         '''
         Convert session contents to be compatible with previous PyMOL versions
@@ -359,6 +394,12 @@ NOTES
         pse_export_version = round(_self.get_setting_float('pse_export_version'), 4)
         legacyscenes = (0 < pse_export_version < 1.76) and _self.get_scene_list()
 
+        if sys.version_info[0] > 2:
+            legacypickle = (0 < pse_export_version < 1.9)
+            if legacypickle:
+                print(' Warning: pse_export_version with Python 3 is experimental')
+            cPickle.configure_legacy_dump(legacypickle)
+
         if legacyscenes:
             _self.pymol._scene_dict = {}
 
@@ -383,7 +424,7 @@ NOTES
                 if ((cache_mode > 0) and (cache_opt != 0)) or (cache_opt==1):
                     _self.cache('optimize')
         for a in _self._pymol._session_save_tasks:
-            if a==None:
+            if a is None:
                 try:
                     _self.lock(_self)
                     r = _cmd.get_session(_self._COb,session,str(names),
@@ -426,9 +467,9 @@ NOTES
                 session = zlib.compress(cPickle.dumps(session, 1))
             return session
         elif _self._raising(r,_self):
-            raise QuietException                  
+            raise QuietException
         return r
-        
+
     def _unit2px(value, dpi, unit=''):
         '''API only. Returns pixel units given a string representation in other units'''
         if cmd.is_string(value):
@@ -501,21 +542,21 @@ PYMOL API
         if format == 'png':
             format = 0
 
-        if prior: 
+        if prior:
             # fetch the prior image, without doing any work (fast-path / non-GLUT thread-safe)
             r = _self._png(str(filename),0,0,float(dpi),0,int(quiet),1,
-                           int(format),_self)            
+                           int(format),_self)
             if r != 1: # no prior image available -- revert to default behavior
                 if prior < 0: # default is to fall back to actual rendering
                     prior = 0
         if not prior:
             dpi = float(dpi)
             if dpi < 0:
-                dpi = cmd.get_setting_float('image_dots_per_inch')
+                dpi = _self.get_setting_float('image_dots_per_inch')
             width = _unit2px(width, dpi)
             height = _unit2px(height, dpi)
 
-            if thread.get_ident() == pymol.glutThread:
+            if _self.is_gui_thread():
                 r = _self._png(str(filename),int(width),int(height),float(dpi),
                                int(ray),int(quiet),0,int(format),_self)
             else:
@@ -525,18 +566,8 @@ PYMOL API
         if _self._raising(r,_self): raise QuietException
         return r
 
-    def export_coords(obj,state,_self=cmd): # experimental
-        r = DEFAULT_ERROR
-        try:
-            _self.lock(_self)   
-            r = _cmd.export_coords(_self._COb,str(obj),int(state)-1)
-        finally:
-            _self.unlock(r,_self)
-        if _self._raising(r,_self): raise QuietException
-        return r
-
     def multisave(filename, pattern="all", state=-1,
-                  append=0, format='', quiet=1, _self=cmd): 
+                  append=0, format='', quiet=1, _self=cmd):
         '''
 DESCRIPTION
 
@@ -606,8 +637,21 @@ ARGUMENTS
         '''
 DESCRIPTION
 
+    Like "get_bytes" but return a unicode string.
+        '''
+        assert format not in ('mmtf',), 'binary format, use get_bytes'
+        b = get_bytes(format, selection, state, ref, ref_state, multi, quiet, _self)
+        if b is None:
+            return None
+        return b.decode('utf-8')
+
+    def get_bytes(format, selection='(all)', state=-1, ref='',
+             ref_state=-1, multi=-1, quiet=1, _self=cmd):
+        '''
+DESCRIPTION
+
     API-only function which exports the selection to a molecular file
-    format and returns it as a string.
+    format and returns it as a binary ("bytes") string.
 
 ARGUMENTS
 
@@ -629,6 +673,85 @@ ARGUMENTS
                     int(state) - 1, str(ref), int(ref_state),
                     int(multi), int(quiet))
 
+    if sys.version_info[0] == 2:
+        get_str = get_bytes
+
+    def multifilesave(filename, selection='*', state=-1, format='', ref='',
+             ref_state=-1, quiet=1, _self=cmd):
+        '''
+DESCRIPTION
+
+    For a selection that spans multiple molecular objects and/or states,
+    save each object and/or state to a separate file. Takes a filename
+    argument with placeholders:
+
+    {name}  : object name
+    {state} : state number
+    {title} : state title
+    {num}   : file number
+    {}      : object name (first) or state (second)
+
+EXAMPLES
+
+    multifilesave /tmp/{name}.pdb
+    multifilesave /tmp/{name}-{state}.cif, state=0
+    multifilesave /tmp/{}-{}.cif, state=0
+    multifilesave /tmp/{}-{title}.sdf, state=0
+        '''
+        for (fname, osele, ostate) in multifilenamegen(
+                filename, selection, int(state), _self):
+            r = _self.save(fname, osele, ostate, format, ref, ref_state, quiet)
+        return r
+
+
+    def multifilenamegen(filename, selection, state, _self=cmd):
+        '''Given a filename pattern, atom selection and state argument,
+        Generate object-state specific filenames and selections.
+        '''
+        import string
+
+        filename = _self.exp_path(filename)
+
+        fmt_keys = [v[1]
+                for v in string.Formatter().parse(filename)
+                if v[1] is not None]
+
+        nindexed = fmt_keys.count('')
+        multiobject = nindexed > 0 or 'name' in fmt_keys or 'num' in fmt_keys
+        multistate = nindexed > 1 or 'state' in fmt_keys or 'title' in fmt_keys
+
+        if not (multiobject or multistate):
+            raise ValueError('need one or more of {name}, {num}, {state}, {title}')
+
+        odata = []
+        for oname in _self.get_object_list(selection):
+            osele = '(%s) & ?%s' % (selection, oname)
+            first = last = state
+
+            if multistate:
+                if state < 0:
+                    first = last = _self.get_object_state(oname)
+
+                if first == 0:
+                    first = 1
+                    last = _self.count_states('%' + oname)
+
+            for ostate in range(first, last + 1):
+                odata.append((oname, osele, ostate))
+
+        # pad {state} and {num} with zeros
+        swidth = len(str(max(v[2] for v in odata)))
+        nwidth = len(str(len(odata)))
+        filename = filename.replace('{state}', '{state:0%d}' % swidth)
+        filename = filename.replace('{num}', '{num:0%d}' % nwidth)
+
+        for num, (oname, osele, ostate) in enumerate(odata, 1):
+            fname = filename.format(oname, ostate,
+                    name=oname, state=ostate, num=num,
+                    title=_self.get_title(oname, ostate))
+            yield fname, osele, ostate
+
+
     def save(filename, selection='(all)', state=-1, format='', ref='',
              ref_state=-1, quiet=1, partial=0,_self=cmd):
         '''
@@ -676,11 +799,11 @@ SEE ALSO
 
         # preprocess selection
         selection = selector.process(selection)
-        #   
+        #
         r = DEFAULT_ERROR
 
         # analyze filename
-        from pymol.importing import filename_to_format
+        from pymol.importing import filename_to_format, _eval_func
         _, _, format_guessed, zipped = filename_to_format(filename)
         filename = _self.exp_path(filename)
 
@@ -709,6 +832,7 @@ SEE ALSO
         if format in savefunctions:
             # generic forwarding to format specific save functions
             func = savefunctions[format]
+            func = _eval_func(func)
             kw = {
                 'filename': filename,
                 'selection': selection,
@@ -763,7 +887,7 @@ SEE ALSO
             with fopen(filename, 'wb') as handle:
                 handle.write(contents)
             r = DEFAULT_SUCCESS
-            
+
         if _self._raising(r,_self): raise QuietException
 
         if not quiet:
@@ -806,23 +930,25 @@ SEE ALSO
     def get_maestr(selection, state=-1, ref='', ref_state=-1, quiet=1, _self=cmd):
         return get_str('mae', selection, state, ref, ref_state, -1, quiet, _self)
 
-    def get_ccp4str(name, state=1, quiet=1, _self=cmd):
+    def get_ccp4str(name, state=1, quiet=1, format='ccp4', _self=cmd):
+        ftype = getattr(loadable, format, -1)
         with _self.lockcm:
             return _cmd.get_ccp4str(_self._COb, str(name),
-                    int(state) - 1, int(quiet))
+                    int(state) - 1, int(quiet), ftype)
 
     def get_psestr(selection, partial, quiet, _self):
         if '(' in selection: # ignore selections
             selection = ''
-        return cPickle.dumps(_self.get_session(selection, partial, quiet), 1)
+        session = _self.get_session(selection, partial, quiet)
+        return cPickle.dumps(session, 1)
 
     def _get_mtl_obj(format, _self):
         # TODO mtl not implemented, always returns empty string
+        if format == 'mtl':
+            raise pymol.CmdException('.MTL export not implemented')
         i = {'mtl': 0, 'obj': 1}.get(format)
         return _self.get_mtl_obj()[i]
 
-    from . import querying
-
     savefunctions = {
         'cif': get_str, # mmCIF
         'xyz': get_str,
@@ -832,6 +958,7 @@ SEE ALSO
         'mol2': get_str,
         'mae': get_str,
         'mol': get_str,
+        'mmtf': get_bytes,
 
         'pse': get_psestr,
         'psw': get_psestr,
@@ -839,14 +966,18 @@ SEE ALSO
         'fasta': get_fastastr,
         'aln': get_alnstr,
         'ccp4': get_ccp4str,
+        'mrc': get_ccp4str,
+        'map': get_ccp4str,
 
         'png': png,
 
         # no arguments (some have a "version" argument)
-        'dae': querying.get_collada,
-        'wrl': querying.get_vrml,
-        'pov': querying.get_povray,
-        'idtf': querying.get_idtf,
+        'dae': 'pymol.querying:get_collada',
+        'gltf': 'pymol.querying:get_gltf',
+        'wrl': 'pymol.querying:get_vrml',
+        'pov': 'pymol.querying:get_povray',
+        'idtf': 'pymol.querying:get_idtf',
         'mtl': _get_mtl_obj, # TODO not implemented
         'obj': _get_mtl_obj,
+        'stl': 'pymol.lazyio:get_stlstr',
     }
